
#include "ddxutils.h"

int DDXFade(LPDIRECTDRAWPALETTE lpDDPal, COLORREF colour, int tickcount, int sizepal)
{
	int i, fine, endtick, tick, nloop, ddRet;
	BYTE red, green, blue;
	PALETTEENTRY pe[256];

	red   = GetRValue(colour);
	green = GetGValue(colour);
	blue  = GetBValue(colour);

	if (sizepal > 16)
		nloop = 256;
	else
		nloop = 16;

	fine = 1;
	while (fine){

		ddRet   = IDirectDrawPalette_GetEntries(lpDDPal, 0, 0, nloop, pe);
		endtick = GetTickCount() + tickcount;
		fine = 0;

		for (i = 0; i<nloop; i++){
			if (pe[i].peRed   > red) { pe[i].peRed--; fine = i; }
			if (pe[i].peRed   < red) { pe[i].peRed++; fine = i; }
			if (pe[i].peGreen > green){ pe[i].peGreen--; fine = i; }
			if (pe[i].peGreen < green){ pe[i].peGreen++; fine = i; }
			if (pe[i].peBlue  > blue) { pe[i].peBlue--; fine = i; }
			if (pe[i].peBlue  < blue) { pe[i].peBlue++; fine = i; }
		}

		while (tick < endtick){
			tick = GetTickCount();
		}

		ddRet = IDirectDrawPalette_SetEntries(lpDDPal, 0, 0, nloop, pe);

	}

	return fine;

}

int DDXCreateSurfaceFromBitmap(LPDIRECTDRAW7 lpDD, char * fName, LPDIRECTDRAWSURFACE7 * lplpSurface)
{
	int  ddret;
	HBITMAP hBmp;
	BITMAP  bm;
	DDSURFACEDESC2 ddsd;

	if (lpDD == NULL || fName == NULL || lplpSurface == NULL)
		return E_INVALIDARG;

	*lplpSurface = 0;

	//  Try to load the bitmap
	hBmp = (HBITMAP)LoadImage(NULL, fName, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);

	if (hBmp == NULL)
		return DDERR_LOADNOBMP;

    // Get size of the bitmap
    GetObject(hBmp, sizeof(bm), &bm);

    // Create a DirectDrawSurface for this bitmap
    memset(&ddsd, 0, sizeof(ddsd));
    ddsd.dwSize         = sizeof(ddsd);
    ddsd.dwFlags        = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
    ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
    ddsd.dwWidth        = bm.bmWidth;
    ddsd.dwHeight       = bm.bmHeight;

	ddret = IDirectDraw7_CreateSurface(lpDD, &ddsd, lplpSurface, NULL);
	if ( ddret != DD_OK )
		return ddret;

    // Draw the bitmap on this surface
	ddret = DDXCopyBitmapToSurface(*lplpSurface, hBmp, 0, 0, 0, 0);
	if (ddret != DD_OK)
		return ddret;

	return DD_OK;
}

int DDXColorMatch(LPDIRECTDRAWSURFACE7 lpDDS, COLORREF colrgb)
{
	int rgbT;
	HDC hdc;
	int dw = CLR_INVALID;
	DDSURFACEDESC2 ddsd;
	int hres;

	//=======================================================================
	// Use GDI SetPixel to color match for us
	//=======================================================================
	if ( colrgb != CLR_INVALID && IDirectDrawSurface7_GetDC(lpDDS, &hdc) == DD_OK ){
		rgbT = GetPixel(hdc, 0, 0);             // save current pixel value
		SetPixel(hdc, 0, 0, colrgb);           // set our value
		IDirectDrawSurface7_ReleaseDC(lpDDS, hdc);
	}

	//=======================================================================
	// Now lock the surface so we can read back the converted color
	//=======================================================================
	ddsd.dwSize = sizeof(ddsd);
	hres = IDirectDrawSurface7_Lock(lpDDS, NULL, &ddsd, 0, NULL);

	while (hres == DDERR_WASSTILLDRAWING){
		hres = IDirectDrawSurface7_Lock(lpDDS, NULL, &ddsd, 0, NULL);
	}

	if (hres == DD_OK){
		dw  = *(DWORD *)ddsd.lpSurface;                  // get DWORD

		if (ddsd.ddpfPixelFormat.dwRGBBitCount < 32){
			dw &= (1 << ddsd.ddpfPixelFormat.dwRGBBitCount) - 1;  // mask it to bpp
		}
		IDirectDrawSurface7_Unlock(lpDDS, NULL);
	}

	//=======================================================================
	// Put the color that was there back.
	//=======================================================================
	if (colrgb != CLR_INVALID && IDirectDrawSurface7_GetDC(lpDDS, &hdc) == DD_OK){
		SetPixel(hdc, 0, 0, rgbT);
		IDirectDrawSurface7_ReleaseDC(lpDDS, hdc);
	}
  	return dw;
}

int DDXSetColorKey(LPDIRECTDRAWSURFACE7 lpDDS, COLORREF colrgb)
{
	DDCOLORKEY ddck;
	ddck.dwColorSpaceLowValue  = DDXColorMatch(lpDDS, colrgb);
	ddck.dwColorSpaceHighValue = ddck.dwColorSpaceLowValue;
	return IDirectDrawSurface7_SetColorKey(lpDDS, DDCKEY_SRCBLT, &ddck);
}

int DDXCopyBitmapToSurface(LPDIRECTDRAWSURFACE7 lpDDS, HBITMAP hbm, int x, int y, int dx, int dy)
{
	HDC hdcImage;
	HDC hdc = 0;
	BITMAP bm;
	DDSURFACEDESC2 ddsd;
	int hr;

	if (hbm == NULL || lpDDS == NULL )
		return 2;

	//=====================================================================
	// Make sure this surface is restored.
	//=====================================================================
	IDirectDrawSurface7_Restore(lpDDS);

	//=====================================================================
	// Select bitmap into a memoryDC so we can use it.
	//=====================================================================
	hdcImage = CreateCompatibleDC(NULL);
	SelectObject(hdcImage, hbm);

	//=====================================================================
	// Get size of the bitmap
	//=====================================================================
	GetObject(hbm, sizeof(bm), &bm);    // Get size of bitmap
	if( dx == 0 ) dx = bm.bmWidth;      // Use the passed size, unless zero
	if( dy == 0 ) dy = bm.bmHeight;

	//=====================================================================
	// Get size of surface.
	//=====================================================================
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_HEIGHT | DDSD_WIDTH;
	IDirectDrawSurface7_GetSurfaceDesc(lpDDS, &ddsd);

	hr = IDirectDrawSurface7_GetDC(lpDDS, &hdc);
	if ( hr == DD_OK ){
		StretchBlt(hdc, x, y, ddsd.dwWidth, ddsd.dwHeight, hdcImage, x, y, dx, dy, SRCCOPY);
		IDirectDrawSurface7_ReleaseDC(lpDDS, hdc);
	}
	DeleteDC(hdcImage);

	return hr;
}

int DDXLoadBmpToSurface(LPDIRECTDRAWSURFACE7 lpDDS, char * fName)
{
	HBITMAP hBmp = NULL;
	int hr;

	hBmp = (HBITMAP)LoadImage(NULL, fName, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE | LR_CREATEDIBSECTION);

	if (hBmp == NULL)
		return DDERR_LOADNOBMP;

	hr = DDXCopyBitmapToSurface(lpDDS, hBmp, 0, 0, 0, 0);

	return hr;
}

DWORD DDXLoadBmpPalette(LPDIRECTDRAW7 lpDD, DWORD flags, char * filename, LPDIRECTDRAWPALETTE * lplpPalette)
{
	int i, n;
	unsigned char temp;
	PALETTEENTRY pe[256];
	BITMAPFILEHEADER bf;
	BITMAPINFOHEADER bi;
	DWORD ddRet;

	FILE * f;

	f = fopen(filename, "rb");
	if(!f)
		return DDERR_PALNOBMP;

	fread(&bf, 1, sizeof(bf), f);

	fread(&bi, 1, sizeof(bi), f);

	fread(pe, 1, sizeof(pe), f);

	fclose(f);

	if (bi.biSize != sizeof(BITMAPINFOHEADER))
		n = 0;
	else if (bi.biBitCount > 8)
		n = 0;
	else if (bi.biClrUsed == 0)
		n = (1 << bi.biBitCount);
	else
		n = bi.biClrUsed;

	//=====================================================================
	// A DIB color table has its colors stored BGR not RGB
	// so flip them around.
	//=====================================================================
	for(int i = 0; i<n; i++){
		temp = pe[i].peRed;
	    pe[i].peRed  = pe[i].peBlue;
	    pe[i].peBlue = temp;
	}

    ddRet = IDirectDraw7_CreatePalette(lpDD, flags, pe, lplpPalette, NULL);
	if(ddRet == DD_OK)
    	return ddRet;
	else
    	return DDERR_PALNOBMP;

}

char * DDXErrorString(DWORD code)
{
	switch( code ){
	case DDERR_PALNOBMP:
		return  "No Palette found.";

	case E_POINTER:
		return  "Indicates that a pointer was invalid.";

	case REGDB_E_CLASSNOTREG:
		return  "Indicates that a specified class is not registered in the registration database.";

	case E_UNEXPECTED:
		return  "Indicates an unexpected error occurred.";

	case CLASS_E_NOAGGREGATION:
		return  "Indicates this class cannot be created as part of an aggregate.";

	case E_NOINTERFACE:
		return  "The requested Interface is not recognised!";

	case DDERR_ALREADYINITIALIZED:
		return  "This object is already initialized";

	case DDERR_CANNOTATTACHSURFACE:
		return  "This surface can not be attached to the requested surface";

	case DDERR_CURRENTLYNOTAVAIL:
		return  "Support is currently not available";

	case DDERR_EXCEPTION:
		return  "An exception was encountered while performing the requested operation";

	case DDERR_GENERIC:
		return  "Generic failure";

	case DDERR_HEIGHTALIGN:
		return  "Height of rectangle not a multiple of required alignment";

	case DDERR_INCOMPATIBLEPRIMARY:
		return  "Unable to match primary surface creation request with existing primary surface";

	case DDERR_INVALIDCAPS:
		return  "One or more of the caps bits passed to the callback are incorrect";

	case DDERR_INVALIDCLIPLIST:
		return  "DirectDraw does not support provided Cliplist";

	case DDERR_INVALIDMODE:
		return  "DirectDraw does not support the requested mode";

	case DDERR_INVALIDOBJECT:
		return  "DirectDraw received a pointer that was an invalid DIRECTDRAW object";

	case DDERR_INVALIDPARAMS:
		return  "Parameters passed to the callback function are incorrect";

	case DDERR_INVALIDPIXELFORMAT:
		return  "Pixel format was invalid as specified";

	case DDERR_INVALIDRECT:
		return  "Rectangle provided was invalid";

	case DDERR_LOCKEDSURFACES:
		return  "One or more surfaces are locked";

	case DDERR_NO3D:
		return  "There is no 3D present";

	case DDERR_NOALPHAHW:
		return  "Operation could not be carried out because there is no alpha accleration hardware present or available";

	case DDERR_NOSTEREOHARDWARE:
		return  "Operation could not be carried out because there is no stereo hardware present or available.";

	case DDERR_NOSURFACELEFT:
		return  "Operation could not be carried out because there is no hardware present which supports stereo surfaces.";

	case DDERR_NOCLIPLIST:
		return  "No clip list available";

	case DDERR_NOCOLORCONVHW:
		return  "There is no color conversion hardware present or available";

	case DDERR_NOCOOPERATIVELEVELSET:
		return  "Create function called without DirectDraw object method SetCooperativeLevel being called";

	case DDERR_NOCOLORKEY:
		return  "Surface doesn't currently have a color key";

	case DDERR_NOCOLORKEYHW:
		return  "There is no hardware supportn of the dest color key";

	case DDERR_NODIRECTDRAWSUPPORT:
		return  "No DirectDraw support possible with current display driver";

	case DDERR_NOEXCLUSIVEMODE:
		return  "Need exclusive mode but the application does not have exclusive mode";

	case DDERR_NOFLIPHW:
		return  "Flipping visible surfaces is not supported";

	case DDERR_NOGDI:
		return  "There is no GDI present";

	case DDERR_NOMIRRORHW:
		return  "Could not be carried out because there is no hardware present or available";

	case DDERR_NOTFOUND:
		return  "Requested item was not found";

	case DDERR_NOOVERLAYHW:
		return  "There is no overlay hardware present or available";

	case DDERR_OVERLAPPINGRECTS:
		return  "Operation could not be carried out because the source and destination rectangles are on the same surface and overlap each other.";

	case DDERR_NORASTEROPHW:
		return  "There is no appropriate raster op hardware present or available";

	case DDERR_NOROTATIONHW:
		return  "There is no rotation hardware present or available";

	case DDERR_NOSTRETCHHW:
		return  "There is no hardware support for stretching";

	case DDERR_NOT4BITCOLOR:
		return  "DirectDrawSurface is not in 4 bit color palette and the requested operation requires 4 bit color palette";

	case DDERR_NOT4BITCOLORINDEX:
		return  "DirectDrawSurface is not in 4 bit color index palette and the requested operation requires 4 bit color index palette";

	case DDERR_NOT8BITCOLOR:
		return  "DirectDraw Surface is not in 8 bit color mode and the requested operation requires 8 bit color";

	case DDERR_NOTEXTUREHW:
		return  "Operation could not be carried out because there is no texture mapping hardware present or available";

	case DDERR_NOVSYNCHW:
		return  "Operation could not be carried out because there is no hardware support for vertical blank synchronized operations";

	case DDERR_NOZBUFFERHW:
		return  "Operation could not be carried out because there is no hardware support for zbuffer blting";

	case DDERR_NOZOVERLAYHW:
		return  "Overlay surfaces could not be z layered based on their BltOrder because the hardware does not support z layering of overlays";

	case DDERR_OUTOFCAPS:
		return  "The hardware needed for the requested operation has already been allocated";

	case DDERR_OUTOFMEMORY:
		return  "DirectDraw does not have enough memory to perform the operation";

	case DDERR_OUTOFVIDEOMEMORY:
		return  "DirectDraw does not have enough memory to perform the operation";

	case DDERR_OVERLAYCANTCLIP:
		return  "Hardware does not support clipped overlays";

	case DDERR_OVERLAYCOLORKEYONLYONEACTIVE:
		return  "Can only have ony color key active at one time for overlays";

	case DDERR_PALETTEBUSY:
		return  "Access to this palette is being refused because the palette is already locked by another thread";

	case DDERR_COLORKEYNOTSET:
		return  "No src color key specified for this operation";

	case DDERR_SURFACEALREADYATTACHED:
		return  "This surface is already attached to the surface it is being attached to";

	case DDERR_SURFACEALREADYDEPENDENT:
		return  "This surface is already a dependency of the surface it is being made a dependency of";

	case DDERR_SURFACEBUSY:
		return  "Access to this surface is being refused because the surface is already locked by another thread";

	case DDERR_CANTLOCKSURFACE:
		return  "Access to this surface is being refused because no driver exists which can supply a pointer to the surface. This is most likely to happen when attempting to lock the primary surface when no DCI provider is present. Will also happen on attempts to lock an optimized surface";

	case DDERR_SURFACEISOBSCURED:
		return  "Access to Surface refused because Surface is obscured";

	case DDERR_SURFACELOST:
		return  "Access to this surface is being refused because the surface is gone. The DIRECTDRAWSURFACE object representing this surface should have Restore called on it";

	case DDERR_SURFACENOTATTACHED:
		return  "The requested surface is not attached";

	case DDERR_TOOBIGHEIGHT:
		return  "Height requested by DirectDraw is too large";

	case DDERR_TOOBIGSIZE:
		return  "Size requested by DirectDraw is too large; The individual height and width are OK";

	case DDERR_TOOBIGWIDTH:
		return  "Width requested by DirectDraw is too large";

	case DDERR_UNSUPPORTED:
		return  "Action not supported";

	case DDERR_UNSUPPORTEDFORMAT:
		return  "FOURCC format requested is unsupported by DirectDraw";

	case DDERR_UNSUPPORTEDMASK:
		return  "Bitmask in the pixel format requested is unsupported by DirectDraw";

	case DDERR_INVALIDSTREAM:
		return  "The specified stream contains invalid data.";

	case DDERR_VERTICALBLANKINPROGRESS:
		return  "Vertical blank is in progress";

	case DDERR_WASSTILLDRAWING:
		return  "Informs DirectDraw that the previous Blt which is transfering information to or from this Surface is incomplete";

	case DDERR_DDSCAPSCOMPLEXREQUIRED:
		return  "The specified surface type requires specification of the COMPLEX flag.";

	case DDERR_XALIGN:
		return  "Rectangle provided was not horizontally aligned on reqd. boundary";

	case DDERR_INVALIDDIRECTDRAWGUID:
		return  "The GUID passed to DirectDrawCreate is not a valid DirectDraw driver identifier";

	case DDERR_DIRECTDRAWALREADYCREATED:
		return  "A DirectDraw object representing this driver has already been created for this process";

	case DDERR_NODIRECTDRAWHW:
		return  "A hardware only DirectDraw object creation was attempted but the driver did not support any hardware";

	case DDERR_PRIMARYSURFACEALREADYEXISTS:
		return  "This process already has created a primary surface";

	case DDERR_NOEMULATION:
		return  "Software emulation not available";

	case DDERR_REGIONTOOSMALL:
		return  "Region passed to Clipper - GetClipList is too small";

	case DDERR_CLIPPERISUSINGHWND:
		return  "An attempt was made to set a clip list for a clipper objec that is already monitoring an hwnd";

	case DDERR_NOCLIPPERATTACHED:
		return  "No clipper object attached to surface object";

	case DDERR_NOHWND:
		return  "Clipper notification requires an HWND or no HWND has previously been set as the CooperativeLevel HWND";

	case DDERR_HWNDSUBCLASSED:
		return  "HWND used by DirectDraw CooperativeLevel has been subclassed, this prevents DirectDraw from restoring state";

	case DDERR_HWNDALREADYSET:
		return  "The CooperativeLevel HWND has already been set. It can not be reset while the process has surfaces or palettes created";

	case DDERR_NOPALETTEATTACHED:
		return  "No palette object attached to this surface";

	case DDERR_NOPALETTEHW:
		return  "No hardware support for 16 or 256 color palettes";

	case DDERR_BLTFASTCANTCLIP:
		return  "If a clipper object is attached to the source surface passed into a BltFast call";

	case DDERR_NOBLTHW:
		return  "No blter";

	case DDERR_NODDROPSHW:
		return  "No DirectDraw ROP hardware";

	case DDERR_OVERLAYNOTVISIBLE:
		return  "Returned when GetOverlayPosition is called on a hidden overlay";

	case DDERR_NOOVERLAYDEST:
		return  "Returned when GetOverlayPosition is called on a overlay that UpdateOverlay has never been called on to establish a destionation";

	case DDERR_INVALIDPOSITION:
		return  "Returned when the position of the overlay on the destionation is no longer legal for that destionation";

	case DDERR_NOTAOVERLAYSURFACE:
		return  "Returned when an overlay member is called for a non-overlay surface";

	case DDERR_EXCLUSIVEMODEALREADYSET:
		return  "An attempt was made to set the cooperative level when it was already set to exclusive";

	case DDERR_NOTFLIPPABLE:
		return  "An attempt has been made to flip a surface that is not flippable";

	case DDERR_CANTDUPLICATE:
		return  "Can't duplicate primary & 3D surfaces, or surfaces that are implicitly created";

	case DDERR_NOTLOCKED:
		return  "Surface was not locked.  An attempt to unlock a surface that was not locked at all, or by this process, has been attempted";

	case DDERR_CANTCREATEDC:
		return  "Windows can not create any more DC's";

	case DDERR_NODC:
		return  "No DC was ever created for this surface";

	case DDERR_WRONGMODE:
		return  "This surface can not be restored because it was created in a different mode";

	case DDERR_IMPLICITLYCREATED:
		return  "This surface can not be restored because it is an implicitly created surface";

	case DDERR_NOTPALETTIZED:
		return  "The surface being used is not a palette-based surface";

	case DDERR_UNSUPPORTEDMODE:
		return  "The display is currently in an unsupported mode";

	case DDERR_NOMIPMAPHW:
		return  "Operation could not be carried out because there is no mip-map texture mapping hardware present or available";

	case DDERR_INVALIDSURFACETYPE:
		return  "The requested action could not be performed because the surface was of the wrong type";

	case DDERR_NOOPTIMIZEHW:
		return  "Device does not support optimized surfaces, therefore no video memory optimized surfaces";

	case DDERR_NOTLOADED:
		return  "Surface is an optimized surface, but has not yet been allocated any memory";

	case DDERR_NOFOCUSWINDOW:
		return  "Attempt was made to create or set a device window without first setting the focus window";

	case DDERR_NOTONMIPMAPSUBLEVEL:
		return  "Attempt was made to set a palette on a mipmap sublevel.";

	case DDERR_DCALREADYCREATED:
		return  "A DC has already been returned for this surface. Only one DC can be retrieved per surface";

	case DDERR_NONONLOCALVIDMEM:
		return  "An attempt was made to allocate non-local video memory from a device that does not support non-local video memory";

	case DDERR_CANTPAGELOCK:
		return  "The attempt to page lock a surface failed";

	case DDERR_CANTPAGEUNLOCK:
		return  "The attempt to page unlock a surface failed";

	case DDERR_NOTPAGELOCKED:
		return  "An attempt was made to page unlock a surface with no outstanding page locks";

	case DDERR_MOREDATA:
		return  "There is more data available than the specified buffer size could hold";

	case DDERR_EXPIRED:
		return  "The data has expired and is therefore no longer valid.";

	case DDERR_TESTFINISHED:
		return  "The mode test has finished executing.";

	case DDERR_NEWMODE:
		return  "The mode test has switched to a new mode.";

	case DDERR_D3DNOTINITIALIZED:
		return  "D3D has not yet been initialized.";

	case DDERR_VIDEONOTACTIVE:
		return  "The video port is not active.";

	case DDERR_NOMONITORINFORMATION:
		return  "The monitor does not have EDID data.";

	case DDERR_NODRIVERSUPPORT:
		return  "The driver does not enumerate display mode refresh rates.";

	case DDERR_DEVICEDOESNTOWNSURFACE:
		return  "Surfaces created by one direct draw device cannot be used directly by another direct draw device";

	case DDERR_NOTINITIALIZED:
		return  "An attempt was made to invoke an interface member of a DirectDraw object created by CoCreateInstance() before it was initialized";

	default:
		return  "Unknown Error";
	}

}



