
#include <windowsx.h>
#include "ddxutils.h"
#include "ddrawres.h"
//
// Uses the DX7 interface
//

HINSTANCE hInst;	// Instance handle
HWND hwndMain;		// Main window handle

LPDIRECTDRAW7 		 lpDD; 			// DirectDraw7 object
LPDIRECTDRAWSURFACE7 lpDDSPrimary; 	// DirectDraw primary surface
LPDIRECTDRAWSURFACE7 lpDDSBack;		// DirectDraw back surface
LPDIRECTDRAWPALETTE  lpDDPal;		// DirectDraw palette
LPDIRECTDRAWSURFACE7 lpDDSOne; 		// Offscreen surface 1
BOOL 				 bActive = TRUE;	// Is App active

DWORD ddRet;   // Return value from calls to Direct Draw

DDSURFACEDESC2  ddsd;		// Surface Description Struct
DDSCAPS2		ddscaps;    // Capabilities of surface Struct
RECT			bkRect;

#define DDERR_PALNOBMP 2
// Define arrays etc for sprites
#define NUMSPRITES  40 // sprites

int mDelay[NUMSPRITES];
int xPos[NUMSPRITES];
int yPos[NUMSPRITES];
int xDir[NUMSPRITES];
int yDir[NUMSPRITES];
int LastTickCount[NUMSPRITES];
int CurrentFrame[NUMSPRITES];
int newTime;
int oldTime;

char * fName = "dd256.bmp";

LRESULT CALLBACK MainWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);

int RestoreAll(void)
{
	int ddRet;

	ddRet = IDirectDrawSurface7_Restore(lpDDSPrimary);
	if (ddRet == DD_OK){
		ddRet = IDirectDrawSurface7_Restore(lpDDSOne);
		if (ddRet == DD_OK)
			DDXLoadBmpToSurface(lpDDSOne, fName);
	}
	return ddRet;
}

void UpdateFrame(void)
{
	int ThisTickCount;
	RECT rcRect;
	HDC hdc = 0;
	int xTemp, yTemp, i;
	static int count;
	static int dCount;
	char s[80];

	while( 1 ){
	    // Blt the main background image
		ddRet = IDirectDrawSurface7_BltFast(lpDDSBack, 0, 0, lpDDSOne, &bkRect, DDBLTFAST_NOCOLORKEY);

		// Display the frames per second
		//================================
		newTime = GetTickCount();
		count ++;
		sprintf(s, "Fms/Sec %d", dCount);

		ddRet = IDirectDrawSurface7_GetDC(lpDDSBack, &hdc);
		if (ddRet == DD_OK){
			SetTextColor(hdc, RGB(255, 255, 255));
			SetBkMode(hdc, TRANSPARENT);
			TextOut(hdc, 0, 0, s, strlen(s));
			IDirectDrawSurface7_ReleaseDC(lpDDSBack, hdc);
		}

		if (newTime > oldTime + 1000){
			oldTime = newTime;
			dCount = count;
			count = 0;
		}
		//================================

		if (ddRet == DD_OK)
			break;

		if (ddRet == DDERR_SURFACELOST){
	      ddRet = RestoreAll();
	      if (ddRet != DD_OK)
			return;
		}

		if (ddRet != DDERR_WASSTILLDRAWING)
      		return;
	}

	if (ddRet != DD_OK)
		return;

	// Do stuff for the donuts

	ThisTickCount = GetTickCount();

	for (i = 0; i<NUMSPRITES; i++){

		// Decide which frame will be blitted next and move sprites
		if ((ThisTickCount - LastTickCount[i]) > mDelay[i]){

			LastTickCount[i] = ThisTickCount;
			// Move to next frame
			CurrentFrame[i]++;

			if (CurrentFrame[i] > 59)
        		CurrentFrame[i] = 0;
		}

		// Fill in rect for position of image to blit to display buffer for each donut
		rcRect.left   = (CurrentFrame[i] % 10) * 64;     // 10 donuts across
		rcRect.top    = ((CurrentFrame[i] / 10) * 64) + 600;
		rcRect.right  = ((CurrentFrame[i] % 10) * 64) + 64;
		rcRect.bottom = ((CurrentFrame[i] / 10) * 64) + 64 + 600;

		// Sort out direction vs. boundary
		if (xPos[i] < -61 || xPos[i] > 636)
			xDir[i] = - xDir[i];

		xPos[i] += xDir[i];

		if (yPos[i] < -61 || yPos[i] > 476)
			yDir[i] = - yDir[i];

		yPos[i] += yDir[i];

		//======================================================================
		// Here's some simple clipping; DDBltFast does not support a ClipperObject.
		// If DDBltFast is asked to blit beyond the display border it refuses.
		// REM these lines and see.
		//======================================================================
		// REM from here
		yTemp = yPos[i];
		if (yTemp < 0){
			rcRect.top -= yTemp;
			yTemp = 0;
		}

		if (yTemp > 416)
			rcRect.bottom -= yTemp - 416;

		xTemp = xPos[i];
		if (xTemp < 0){
			rcRect.left -= xTemp;
			xTemp = 0;
		}

		if (xTemp > 576)
			rcRect.right -= xTemp - 576;

		// To here
		while( 1 ){ // Transparent blit using the source's colour key.

			ddRet = IDirectDrawSurface7_BltFast(lpDDSBack, xTemp, yTemp, lpDDSOne, &rcRect, DDBLTFAST_SRCCOLORKEY);

			if (ddRet == DDERR_INVALIDRECT)
				MessageBeep(16); // Trouble with clipping!?

      		if (ddRet == DD_OK)
				break;

			if (ddRet == DDERR_SURFACELOST){
				ddRet = RestoreAll();
				if (ddRet != DD_OK)
          			return;
			}

			if (ddRet != DDERR_WASSTILLDRAWING)
				return;
		} // end while

	} // end next

	// Lastly flip the front and back surfaces
	while( 1 ){

		ddRet = IDirectDrawSurface7_Flip(lpDDSPrimary, NULL, 0);
		if (ddRet == DD_OK)
			break;

		if (ddRet == DDERR_SURFACELOST){
			ddRet = RestoreAll();
			if (ddRet != DD_OK)
				break;
		}

		if (ddRet != DDERR_WASSTILLDRAWING)
			break;
	}

}

static BOOL InitApplication(void)
{
	WNDCLASS wc;

	memset(&wc,0,sizeof(WNDCLASS));
	wc.style 			= CS_HREDRAW|CS_VREDRAW |CS_DBLCLKS ;
	wc.lpfnWndProc 		= (WNDPROC)MainWndProc;
	wc.hInstance 		= hInst;
	wc.hbrBackground 	= (HBRUSH)(COLOR_WINDOW+1);
	wc.lpszClassName 	= "ddrawWndClass";
	wc.lpszMenuName 	= NULL;
	wc.hCursor 			= LoadCursor(NULL,IDC_ARROW);
	wc.hIcon 			= LoadIcon(NULL,IDI_APPLICATION);
	if (!RegisterClass(&wc))
		return 0;
	return 1;
}

HWND CreateddrawWndClassWnd(void)
{
	return CreateWindow("ddrawWndClass","ddraw",
		WS_POPUP,
		CW_USEDEFAULT,0,CW_USEDEFAULT,0,
		NULL,
		NULL,
		hInst,
		NULL);
}

void MainWndProc_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	switch(id) {
	case IDM_EXIT:
		PostMessage(hwnd,WM_CLOSE,0,0);
		break;
	}
}

LRESULT CALLBACK MainWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	switch (msg) {
	case WM_ACTIVATEAPP:
		bActive = wParam;
		break;

	case WM_KEYDOWN:
	    switch (wParam){
    	case VK_UP:
      		bkRect.top -= 2;
			bkRect.bottom -= 2;

			if (bkRect.top < 0){
		        bkRect.top    = 0;
        		bkRect.bottom = 480;
     		}
		break;

		case VK_DOWN:
  		    bkRect.top    += 2;
  		    bkRect.bottom += 2;

  		    if (bkRect.top > 120){
  		      bkRect.top    = 120;
  		      bkRect.bottom = 600;
			}
		break;

    	case VK_F12: case VK_ESCAPE:
			DDXFade(lpDDPal, RGB(0, 0, 0), 6, 256);
			PostMessage(hwnd, WM_CLOSE, 0, 0);
    	}
    	break;

	case WM_COMMAND:
		HANDLE_WM_COMMAND(hwnd, wParam, lParam, MainWndProc_OnCommand);
		break;
	case WM_DESTROY:
		PostQuitMessage(0);
		break;
	default:
		return DefWindowProc(hwnd, msg, wParam, lParam);
	}
	return 0;
}

void DestroyDDObjects(void)
{
	// After Releasing Objects set Objects to NULL
	if (lpDD != NULL){

		if (lpDDSPrimary != NULL){
			lpDDSPrimary->lpVtbl->Release(lpDDSPrimary);
			lpDDSPrimary = NULL;
		}

		if (lpDDSOne != NULL){
			lpDDSOne->lpVtbl->Release(lpDDSOne) ;
			lpDDSOne = NULL;
		}

		if (lpDDPal != NULL){
			lpDDPal->lpVtbl->Release(lpDDPal) ;
			lpDDPal = NULL;
		}

		lpDD->lpVtbl->Release(lpDD) ;
		lpDD = NULL;
	}
}

void DDInitError(char * s, DWORD code )
{
	//======================================================================
	// For this case (DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN) the MessageBox
	// may not display correctly, dependant on what stage Initialisation reached;
	// it is better to hide the main window before displaying the MessageBox().
	//======================================================================
	DestroyDDObjects();
	ShowWindow(hwndMain, SW_HIDE);
	MessageBox(NULL, DDXErrorString(code), "DirectX Error", MB_OK | MB_ICONEXCLAMATION);
}

int Random(int x)
{
	return (rand()%x);
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, char * lpCmdLine, int nCmdShow)
{
	MSG msg;

	hInst = hInstance;
	if (!InitApplication())
		return 0;

	if ((hwndMain = CreateddrawWndClassWnd()) == (HWND)0)
		return 0;

	// Set initial position, speed, direction and delay for each sprite
	for (int i = 0; i<NUMSPRITES; i++){
		xPos[i] = 2 + Random(574);
		yPos[i] = 2 + Random(410);
		if (i & 1){ // odd

			xDir[i] = 1 + Random(2);
			yDir[i] = 1 + Random(2);
		}
		else{		// even

			xDir[i] = - (1 + Random(2));
			yDir[i] = - (1 + Random(2));
		}
		mDelay[i] = 2 + i;
	}

	// For initial position of background
	bkRect.left   = 0;
	bkRect.top    = 0;
	bkRect.right  = 640;
	bkRect.bottom = 480;

	ddRet = DirectDrawCreateEx(NULL, &lpDD, &IID_IDirectDraw7, NULL);
	if(ddRet != DD_OK){
		DDInitError("DirectDrawCreateEx Failed", ddRet);
		goto Cleanup;
	}

	// Get exclusive fullscreen mode
	ddRet = IDirectDraw7_SetCooperativeLevel(lpDD, hwndMain, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN);
	if ( ddRet != DD_OK ){
		DDInitError("SetCooperativeLevel Failed", ddRet);
		goto Cleanup;
	}

	// Set the video mode to 640x480x8 (256 colours)
	ddRet = IDirectDraw7_SetDisplayMode(lpDD, 640, 480, 8, 0, 0);
	if ( ddRet != DD_OK ) {
		DDInitError("SetDisplayMode Failed", ddRet);
		goto Cleanup;
	}

	// Create the primary surface with 1 back buffer
	memset(&ddsd, 0, sizeof(ddsd));       // Clear structure
	ddsd.dwSize            = sizeof(ddsd);
	ddsd.dwFlags           = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	ddsd.ddsCaps.dwCaps    = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
	ddsd.dwBackBufferCount = 1;

	ddRet = IDirectDraw7_CreateSurface(lpDD, &ddsd, &lpDDSPrimary, NULL);
	if ( ddRet != DD_OK ){
	    DDInitError("CreateSurface Failed", ddRet);
		goto Cleanup;
	}

	// Get the pointer to the back buffer
	ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
	ddRet = IDirectDrawSurface7_GetAttachedSurface(lpDDSPrimary, &ddscaps, &lpDDSBack);
	if (ddRet != DD_OK){
		DDInitError("GetAttachedSurface Failed", ddRet);
		goto Cleanup;
	}

	// Load palette from image. We want all colors to fade so DDPCAPS_ALLOW256
	ddRet = DDXLoadBmpPalette(lpDD, DDPCAPS_8BIT | DDPCAPS_ALLOW256, fName, &lpDDPal);
	if ( ddRet != DD_OK ){
		DDInitError("DDXLoadBmpPalette Failed", ddRet);
		goto Cleanup;
	}

	// Set Palette of primary surface
	ddRet = IDirectDrawSurface7_SetPalette(lpDDSPrimary, lpDDPal);
	if ( ddRet != DD_OK ){
		DDInitError("SetPalette Failed", ddRet);
		goto Cleanup;
	}

	// Create another OFFSCREENPLAIN surface
	memset(&ddsd, 0, sizeof(ddsd));        // Clear structure
	ddsd.dwSize          = sizeof(ddsd);
	ddsd.dwFlags         = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	ddsd.ddsCaps.dwCaps  = DDSCAPS_OFFSCREENPLAIN;
	ddsd.dwWidth         = 640; // Width and
	ddsd.dwHeight        = 984; // Height of Bitmap
	ddRet = IDirectDraw7_CreateSurface(lpDD, &ddsd, &lpDDSOne, NULL);
	if ( ddRet != DD_OK ){
		DDInitError("CreateSurface Failed", ddRet);
		goto Cleanup;
	}

	// Load the image onto the OFFSCREENPLAIN surface
	ddRet = DDXLoadBmpToSurface(lpDDSOne, fName);
	if( ddRet != DD_OK ){
		DDInitError("DDLoadBmpToSurface Failed", ddRet);
		goto Cleanup;
	}

	// Set the color key for this image (black for this example) for sprites Transparent blit.
	DDXSetColorKey(lpDDSOne, RGB(0, 0, 0));

	ShowWindow(hwndMain, SW_SHOW);

	oldTime = GetTickCount();

    while( 1 ) {
        if( PeekMessage( &msg, NULL, 0, 0, PM_NOREMOVE ) ){
            if( !GetMessage( &msg, NULL, 0, 0 ) ) {
                break;
            }
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }else if( bActive ){
            UpdateFrame();
        }else{
            WaitMessage();
        }
    }

Cleanup:
	DestroyDDObjects();

	return 0;
}
